<?php

namespace Drupal\pbs_media_manager;

use DateTime;
use DateTimeZone;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Exception;
use stdClass;

/**
 * Class SeasonManager.
 *
 * @package Drupal\pbs_media_manager
 */
class SeasonManager extends ApiContentManagerBase {

  /**
   * State key for the last update DateTime.
   */
  const LAST_UPDATE_KEY = 'pbs_media_manager.shows.last_update';

  /**
   * {@inheritdoc}
   */
  public static function getEntityTypeId(): string {
    return 'node';
  }

  /**
   * {@inheritdoc}
   */
  public function getBundleId(): string {
    return $this->config->get('seasons.drupal_season_content');
  }

  /**
   * {@inheritdoc}
   */
  public static function getQueueName(): string {
    return 'pbs_media_manager.queue.seasons';
  }

  /**
   * {@inheritdoc}
   */
  public function getLastUpdateTime(): DateTime {
    return $this->state->get(
      self::LAST_UPDATE_KEY,
      new DateTime('@1')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function setLastUpdateTime(DateTime $time): void {
    $this->state->set(self::LAST_UPDATE_KEY, $time);
  }

  /**
   * {@inheritdoc}
   */
  public function resetLastUpdateTime(): void {
    $this->state->delete(self::LAST_UPDATE_KEY);
  }

  /**
   * {@inheritdoc}
   */
  public static function getAutoUpdateConfigName(): string {
    return 'season.queue.autoupdate';
  }

  /**
   * {@inheritdoc}
   */
  public static function getAutoUpdateIntervalConfigName(): string {
    return 'seasons.queue.autoupdate_interval';
  }

  /**
   * {@inheritdoc}
   */
  public function updateQueue(DateTime $since = NULL): bool {
    $dt_start = new DateTime();

    $show_vocabulary = $this->config->get('shows.show_vocabulary');
    if (!empty($show_vocabulary) && $show_vocabulary != 'unused') {
      $taxonomy_manager = $this->entityTypeManager->getStorage('taxonomy_term');
      $subscribed_shows = $taxonomy_manager->loadByProperties([
        'vid' => $show_vocabulary,
      ]);
      foreach ($subscribed_shows as $show) {
        $show_guid = $show->get('field_show_guid')->value;
        $seasons = $this->client->getSeasons($show_guid);
        foreach ($seasons as $season) {
          $season_item = $this->client->getSeason($season->id);
          $this->getQueue()->createItem($season_item);
        }
      }
    }
    else {
      $shows = $this->client->getShows(['fetch-related' => FALSE]);
      foreach ($shows as $show) {
        try {
          $seasons = $this->client->getSeasons($show->id);
        }
        catch (\Exception $e) {
          // If we don't have access to the season, just skip it.
          continue;
        }
        foreach ($seasons as $season) {
          $season_item = $this->client->getSeason($season->id);
          $this->getQueue()->createItem($season_item);
        }
      }
    }

    $this->setLastUpdateTime($dt_start);

    return TRUE;
  }

  /**
   * Gets an API response for a single season.
   *
   * @param string $guid
   *   ID of the season to get.
   *
   * @return object|null
   *   season item data from the API or NULL.
   */
  public function getSeason(string $guid): ?stdClass {
    return $this->client->getSeason($guid);
  }

  /**
   * Get season nodes with optional properties.
   *
   * @param array $properties
   *   Properties to filter season nodes.
   *
   * @return \Drupal\node\NodeInterface[]
   *   seasons nodes.
   */
  public function getSeasonNodes(array $properties = []): array {
    try {
      $definition = $this->entityTypeManager->getDefinition(self::getEntityTypeId());
      $storage = $this->entityTypeManager->getStorage(self::getEntityTypeId());
      $nodes = $storage->loadByProperties([
        $definition->getKey('bundle') => $this->getBundleId(),
      ] + $properties);
    }
    catch (Exception $e) {
      // Let NULL fall through.
      $nodes = [];
    }

    return $nodes;
  }

  /**
   * Adds or updates a season node based on API data.
   *
   * @param object $item
   *   An API response object for the season.
   * @param bool $force
   *   Whether or not to "force" the update. If FALSE (default) and a current
   *   episode content node is found with a matching or newer "last_updated"
   *   value, the existing node will not be updated. If TRUE, "last_updated" is
   *   ignored.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Exception
   */
  public function addOrUpdateSeason(object $item, $force = FALSE): void {

    $mappings = $this->config->get('seasons.mappings');
    $season_id = $item->id;
    $node = $this->getOrCreateNode($item->id, $this->getBundleId(), 'seasons');
    $action = $node->IsNew() ? 'created' : 'updated';
    $attributes = $item->attributes;

    // Confirm that the "Updated at" field is set.
    $updated_at_field = $mappings['updated_at'];
    if (empty($updated_at_field) || $updated_at_field == 'unused') {
      $this->logger->error($this->t('To import seasons, please select an "updated_at" value in the "Season Media Manager field mappings."'));
      return;
    }
    // Get to node update time.
    $drupal_updated_at = $node->get($updated_at_field)->value;
    // If there is a Drupal node and the updated at field is empty,
    // update the data in the node.
    if (empty($drupal_updated_at)) {
      $action == 'updated';
    }
    // For already-existing Drupal nodes, check if the item has changed in PBS.
    if (!empty($drupal_updated_at) && $action == 'updated') {
      // Get the updated time from Media Manager.
      $pbs_updated_at = self::getLatestUpdatedAt($item)->getTimestamp();
      // Get Drupal's updated at time in UTC time.
      $dt_updated = new DrupalDateTime($drupal_updated_at, DateTimeItemInterface::STORAGE_TIMEZONE);
      $drupal_updated_at = $dt_updated->getTimestamp();
      // Stop if $force is FALSE and the episode was not updated in the API.
      if (!$force && $pbs_updated_at <= $drupal_updated_at) {
        $this->logger->notice($this->t('Season @title has not changed', [
          '@title' => $node->getTitle(),
        ]));
        return;
      }
    }

    $season_number = $attributes->ordinal ?: '';
    $show_title = $attributes->show->attributes->title ?: '';
    if (!empty($attributes->title)) {
      $season_title = $attributes->title;
    }
    elseif (!empty($season_number) && !empty($show_title)) {
      $season_title = $show_title . ": Season " . $season_number;
    }
    else {
      $this->logger->error($this->t('The season ID @id is missing a title, season number, and/or associated show title.', [
        '@id' => $season_id,
      ]));
      return;
    }
    $node->setTitle($season_title);

    if (!empty($item->id) && $mappings['id'] != 'unused') {
      $node->set($mappings['id'], $item->id);
    }
    if (!empty($attributes->ordinal) && $mappings['ordinal'] != 'unused') {
      $node->set($mappings['ordinal'], $attributes->ordinal);
    }
    if (!empty($attributes->description_long) &&
      $mappings['description_long'] != 'unused') {
      $node->set($mappings['description_long'], $attributes->description_long);
    }
    if (!empty($attributes->description_short) &&
      $mappings['description_short'] != 'unused') {
      $node->set($mappings['description_short'], $attributes->description_short);
    }

    $updated_at = self::getLatestUpdatedAt($item);
    $node->set(
      $mappings['updated_at'],
      $updated_at->format(DateTimeItemInterface::DATETIME_STORAGE_FORMAT)
    );

    // Add the show reference, if configured.
    $show_ref_field = $this->config->get('seasons.references.show');
    $show_guid = $attributes->show->id;
    $show_bundle = $this->config->get('shows.drupal_show_content');
    if (!empty($show_guid) && !empty($show_bundle) && $show_ref_field != 'unused') {
      $show_node = $this->getNodeByGuid($show_guid, $show_bundle, 'shows');
      if (!empty($show_node) && !empty($show_node->id())) {
        $node->set($show_ref_field, ['target_id' => $show_node->id()]);
      }
    }

    $node->save();

    if (!empty($node->getTitle()) & !empty($action)) {
      $this->logger->notice($this->t('Season @title has been @action', [
        '@title' => $node->getTitle(),
        '@action' => $action,
      ]));
    }
  }

}
